import asyncio
import base64
import json
import os
import shutil
import time
from pathlib import Path
from typing import List, Optional

import aiofiles
import httpx
import polars as pl
from pydantic import UUID4
from sqlmodel import select
from sqlmodel.ext.asyncio.session import AsyncSession

from hudson._env import env
from hudson._types import DataArray, Message
from hudson.models import (
    Dataset,
    DatasetCreate,
    Namespace,
    NamespaceCreate,
    Subscription,
    SubscriptionCreate,
    Topic,
    TopicCreate,
)


class NamespaceService:
    async def get_by_name(self, name: str, psql: AsyncSession) -> Optional[Namespace]:
        results = await psql.execute(select(Namespace).where(Namespace.name == name))
        return results.scalars().first()

    async def get(self, namespace_id: UUID4, psql: AsyncSession) -> Optional[Namespace]:
        results = await psql.execute(
            select(Namespace).where(Namespace.id == namespace_id)
        )
        return results.scalars().first()

    async def create(
        self, namespace_create: NamespaceCreate, psql: AsyncSession
    ) -> Namespace:
        namespace = Namespace(name=namespace_create.name)
        psql.add(namespace)
        await psql.commit()
        await psql.refresh(namespace)
        return namespace

    async def list(self, name: Optional[str], psql: AsyncSession) -> List[Namespace]:
        if name is not None:
            results = (
                (
                    await psql.execute(
                        select(Namespace)
                        .where(Namespace.name.contains(name))  # type: ignore
                        .order_by(Namespace.name)
                    )
                )
                .scalars()
                .all()
            )
        else:
            results = (
                (await psql.execute(select(Namespace).order_by(Namespace.name)))
                .scalars()
                .all()
            )
        return results

    async def delete(self, namespace_id: UUID4, psql: AsyncSession) -> Namespace:
        results = await psql.execute(
            select(Namespace).where(Namespace.id == namespace_id)
        )
        namespace = results.scalars().first()
        if namespace:
            await psql.delete(namespace)
            await psql.commit()

        namespace_dir = Path(env.DATASETS_PATH) / str(namespace.id)
        if namespace_dir.exists():
            shutil.rmtree(namespace_dir)

        return namespace


class TopicsService:
    async def list(
        self, namespace_id: UUID4, name: Optional[str], psql: AsyncSession
    ) -> List[Topic]:
        if name is not None:
            results = (
                (
                    await psql.execute(
                        select(Topic)
                        .where(
                            Topic.namespace_id == namespace_id,
                            Topic.name.contains(name),  # type: ignore
                        )
                        .order_by(Topic.name)
                    )
                )
                .scalars()
                .all()
            )
        else:
            results = (
                (
                    await psql.execute(
                        select(Topic)
                        .where(Topic.namespace_id == namespace_id)
                        .order_by(Topic.name)
                    )
                )
                .scalars()
                .all()
            )
        return results

    async def get(
        self, topic_id: UUID4, namespace_id: UUID4, psql: AsyncSession
    ) -> Optional[Topic]:
        results = await psql.execute(
            select(Topic).where(
                Topic.id == topic_id,
                Topic.namespace_id == namespace_id,
            )
        )
        return results.scalars().first()

    async def create(self, topic_create: TopicCreate, psql: AsyncSession) -> Topic:
        results = await psql.execute(
            select(Topic).where(
                Topic.name == topic_create.name,
                Topic.namespace_id == topic_create.namespace_id,
            )
        )
        topic = results.scalars().first()
        if topic is not None:
            return topic
        topic = Topic(**topic_create.dict())
        psql.add(topic)
        await psql.commit()
        await psql.refresh(topic)
        return topic

    async def delete(
        self, topic_id: UUID4, namespace_id: UUID4, psql: AsyncSession
    ) -> Topic:
        results = await psql.execute(
            select(Topic).where(
                Topic.id == topic_id, Topic.namespace_id == namespace_id
            )
        )
        topic = results.scalars().first()
        if topic:
            await psql.delete(topic)
            await psql.commit()
        return topic

    async def publish_message(self, topic: Topic, message: Message) -> None:
        # TODO: hudson supports http-push publishing to HTTPS endpoints.
        # hudson should support other modes and other protocols in the future
        # e.g. pull, gRPC, carrier pigeon, etc.
        await asyncio.gather(
            *[
                self.publish_message_to_subscription(
                    subscription=sub,
                    message=base64.b64decode(message.data.encode("utf-8")).decode(
                        "utf-8"
                    ),
                )
                for sub in topic.subscriptions
            ]
        )

    async def publish_message_to_subscription(
        self, subscription: Subscription, message: str
    ) -> None:
        async with httpx.AsyncClient() as client:
            await client.post(
                str(subscription.push_endpoint),
                json=message,
                headers={"Content-Type": "application/json"},
            )


class SubscriptionsService:
    async def create(
        self, subscription_create: SubscriptionCreate, psql: AsyncSession
    ) -> Subscription:
        results = await psql.execute(
            select(Subscription).where(
                Subscription.name == subscription_create.name,
                Subscription.topic_id == subscription_create.topic_id,
            )
        )
        subscription = results.scalars().first()
        if subscription is not None:
            return subscription
        subscription = Subscription(**subscription_create.dict())
        psql.add(subscription)
        await psql.commit()
        await psql.refresh(subscription)
        return subscription

    async def list(
        self,
        namespace_id: UUID4,
        topic_id: UUID4,
        name: Optional[str],
        psql: AsyncSession,
    ) -> List[Subscription]:
        if name is not None:
            results = (
                (
                    await psql.execute(
                        select(Subscription)
                        .join(Topic, Subscription.topic_id == Topic.id)
                        .where(
                            Topic.namespace_id == namespace_id,
                            Subscription.topic_id == topic_id,
                            Subscription.name.contains(name),  # type: ignore
                        )
                        .order_by(Subscription.name)
                    )
                )
                .scalars()
                .all()
            )
        else:
            results = (
                (
                    await psql.execute(
                        select(Subscription)
                        .join(Topic, Subscription.topic_id == Topic.id)
                        .where(
                            Topic.namespace_id == namespace_id,
                            Subscription.topic_id == topic_id,
                        )
                        .order_by(Subscription.name)
                    )
                )
                .scalars()
                .all()
            )
        return results

    async def get(
        self,
        subscription_id: UUID4,
        topic_id: UUID4,
        namespace_id: UUID4,
        psql: AsyncSession,
    ) -> Optional[Subscription]:
        results = await psql.execute(
            select(Subscription)
            .join(Topic, Subscription.topic_id == Topic.id)
            .where(
                Subscription.id == subscription_id,
                Subscription.topic_id == topic_id,
                Topic.namespace_id == namespace_id,
            )
        )
        return results.scalars().first()

    async def delete(
        self,
        subscription_id: UUID4,
        topic_id: UUID4,
        namespace_id: UUID4,
        psql: AsyncSession,
    ) -> Subscription:
        results = await psql.execute(
            select(Subscription)
            .join(Topic, Subscription.topic_id == Topic.id)
            .where(
                Subscription.id == subscription_id,
                Subscription.topic_id == topic_id,
                Topic.namespace_id == namespace_id,
            )
        )
        subscription = results.scalars().first()
        if subscription:
            await psql.delete(subscription)
            await psql.commit()
        return subscription


class DatasetsService:
    async def create(
        self, dataset_create: DatasetCreate, psql: AsyncSession
    ) -> Dataset:
        results = await psql.execute(
            select(Dataset).where(
                Dataset.name == dataset_create.name,
                Dataset.namespace_id == dataset_create.namespace_id,
            )
        )
        dataset = results.scalars().first()
        if dataset is not None:
            return dataset
        dataset = Dataset(**dataset_create.dict())
        psql.add(dataset)
        await psql.commit()
        await psql.refresh(dataset)
        return dataset

    async def create_directory_for_dataset(self, dataset: Dataset) -> None:
        dataset_path = (
            Path(env.DATASETS_PATH) / str(dataset.namespace_id) / str(dataset.id)
        )
        dataset_path.mkdir(parents=True, exist_ok=True)

    async def list(
        self,
        namespace_id: UUID4,
        name: Optional[str],
        psql: AsyncSession,
    ) -> List[Dataset]:
        if name is not None:
            results = (
                (
                    await psql.execute(
                        select(Dataset)
                        .where(
                            Dataset.namespace_id == namespace_id,
                            Dataset.name.contains(name),  # type: ignore
                        )
                        .order_by(Dataset.name)
                    )
                )
                .scalars()
                .all()
            )
        else:
            results = (
                (
                    await psql.execute(
                        select(Dataset)
                        .where(Dataset.namespace_id == namespace_id)
                        .order_by(Dataset.name)
                    )
                )
                .scalars()
                .all()
            )
        return results

    async def get(
        self, dataset_id: UUID4, namespace_id: UUID4, psql: AsyncSession
    ) -> Optional[Dataset]:
        results = await psql.execute(
            select(Dataset).where(
                Dataset.id == dataset_id,
                Dataset.namespace_id == namespace_id,
            )
        )
        return results.scalars().first()

    async def delete(
        self, dataset_id: UUID4, namespace_id: UUID4, psql: AsyncSession
    ) -> Dataset:
        results = await psql.execute(
            select(Dataset).where(
                Dataset.id == dataset_id, Dataset.namespace_id == namespace_id
            )
        )
        dataset = results.scalars().first()
        if dataset:
            await psql.delete(dataset)
            await psql.commit()

        dataset_path = Path(env.DATASETS_PATH) / str(namespace_id) / str(dataset_id)
        if dataset_path.exists():
            shutil.rmtree(dataset_path)

        return dataset

    async def write(
        self, namespace_id: UUID4, dataset_id: UUID4, data_array: DataArray
    ) -> None:
        dataset_path = (
            Path(env.DATASETS_PATH)
            / str(namespace_id)
            / str(dataset_id)
            / f"{int(time.time() * 1000)}.jsonl"
        )
        async with aiofiles.open(dataset_path, "w") as f:
            for d in data_array.data:
                await f.write(json.dumps(d.dict()))
        df = pl.read_ndjson(dataset_path)
        df.write_ipc(dataset_path.with_suffix(".arrow"))
        os.remove(dataset_path)

    async def read(
        self,
        namespace_id: UUID4,
        dataset_id: UUID4,
    ) -> DataArray:
        dataset_path = (
            Path(env.DATASETS_PATH) / str(namespace_id) / str(dataset_id) / "*.arrow"
        )
        df = pl.read_ipc(dataset_path)
        return DataArray(data=df.to_dicts())


namespace_service = NamespaceService()
topics_service = TopicsService()
subscriptions_service = SubscriptionsService()
datasets_service = DatasetsService()



import random
import string
import time

import torch
from docarray import Document, DocumentArray

from hudson import hudson_client

#
#   Create a namespace.
#
print("🎉 Creating a namespace")
namespace = hudson_client.create_namespace(
    name="text-example",
)

#
#   Create a dataset.
#
print("🎉 Creating a dataset")
dataset = hudson_client.create_dataset(
    namespace_id=namespace.id,
    name="text-example-dataset",
)

#
#   Write a DocumentArray to the dataset.
#
print("🧱 Building the dataset")
n = 100
str_len = 100
data = DocumentArray(
    [
        Document(
            text="".join(
                random.choices(string.ascii_uppercase + string.digits, k=str_len)
            ),
            embedding=torch.randn(768),
        )
        for _ in range(n)
    ]
)
print("✍️  Writing to the dataset")
t0 = time.time()
hudson_client.write_dataset(
    namespace_id=namespace.id,
    dataset_id=dataset.id,
    data=data,
)
print(f"⏰ Took {time.time() - t0:.2f} seconds to write {n} documents")


print("📖 Reading from the dataset")
t0 = time.time()
da = hudson_client.read_dataset(
    namespace_id=namespace.id,
    dataset_id=dataset.id,
)
print(f"⏰ Took {time.time() - t0:.2f} seconds to read {n} documents")

print("🧹 Cleaning up!")
hudson_client.delete_namespace(namespace.id)



import base64
import json
import sys
from datetime import datetime
from enum import Enum, unique
from typing import Dict, List, Optional

from docarray.document.pydantic_model import PydanticDocumentArray
from pydantic import BaseModel, Json, StrictInt, StrictStr, validator


@unique
class DeliveryType(str, Enum):
    # PULL = "pull" – TODO: coming soon!
    PUSH = "push"


class HealthResponse(BaseModel):
    message: StrictStr
    version: StrictStr
    time: datetime


class DataArray(BaseModel):
    data: PydanticDocumentArray

    class Config:
        schema_extra = {
            "example": {
                "data": [
                    {
                        "id": "1",
                        "text": "hello",
                        "tags": {"foo": "bar"},
                        "embedding": [0.1, 0.2, 0.3],
                    },
                    {
                        "id": "2",
                        "text": "world",
                        "tags": {"foo": "bar"},
                        "embedding": [0.1, 0.2, 0.3],
                    },
                ]
            }
        }


class Scope(BaseModel):
    type: StrictStr
    asgi: Optional[Dict]
    http_version: StrictStr
    method: StrictStr
    scheme: StrictStr
    root_path: Optional[StrictStr]
    path: StrictStr
    raw_path: Optional[str]
    headers: List
    query_string: bytes


class RequestLoggerMessage(BaseModel):
    scope: Scope
    _stream_consumed: bool
    _is_disconnected: bool


class ResponseLoggerMessage(BaseModel):
    status_code: StrictInt
    body: Json
    raw_headers: List


class Message(BaseModel):
    data: StrictStr

    @validator("data", pre=True, always=True)
    def validate_data_is_less_than_10mb(cls: BaseModel, v: str) -> str:
        if sys.getsizeof(v) > 10_000_000:
            raise ValueError("Message data must be less than or equal to 10MB.")
        try:
            json.loads(base64.b64decode(v.encode("utf-8")).decode("utf-8"))
        except json.JSONDecodeError:
            raise ValueError("Message data must be a valid base64 encoded JSON string.")
        return v

    class Config:
        schema_extra = {
            "example": {
                "data": {"message": "Hello world!"},
            }
        }



import asyncio
import shutil
from typing import AsyncGenerator, Generator

import pytest
import pytest_asyncio
from fastapi import Request
from httpx import AsyncClient
from sqlalchemy.orm import sessionmaker
from sqlmodel import SQLModel
from sqlmodel.ext.asyncio.session import AsyncSession
from typer.testing import CliRunner

from hudson._env import env
from hudson.db import async_psql_engine
from hudson.server.main import app as server_app


@pytest.fixture(scope="function")
def runner() -> Generator:
    runner = CliRunner()
    yield runner


@pytest.fixture(scope="session")
def event_loop(request: Request) -> Generator:
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest_asyncio.fixture(scope="session")
async def client() -> AsyncGenerator:
    async with AsyncClient(
        app=server_app,
        base_url=env.HUDSON_SERVER_URL,
    ) as client:
        yield client


@pytest_asyncio.fixture(scope="function", autouse=True)
async def async_db_session() -> AsyncGenerator:
    session = sessionmaker(
        async_psql_engine, class_=AsyncSession, expire_on_commit=False
    )

    async with session() as s:
        async with async_psql_engine.begin() as conn:
            await conn.run_sync(SQLModel.metadata.create_all)
        yield s

    async with async_psql_engine.begin() as conn:
        await conn.run_sync(SQLModel.metadata.drop_all)

    await async_psql_engine.dispose()


@pytest.fixture(scope="session", autouse=True)
async def delete_test_database() -> AsyncGenerator:
    yield
    shutil.rmtree(env.DATASETS_PATH, ignore_errors=True)



from datetime import datetime
from typing import List, Optional
from uuid import uuid4

from pydantic import UUID4, AnyHttpUrl, BaseModel, validator
from sqlalchemy import Column, DateTime, UniqueConstraint
from sqlmodel import Field, Relationship, SQLModel

from hudson._types import DeliveryType


class TimestampsMixin(BaseModel):
    created_at: Optional[datetime] = Field(
        sa_column=Column(
            DateTime,
            default=datetime.utcnow,
            nullable=False,
        )
    )
    updated_at: Optional[datetime] = Field(
        sa_column=Column(
            DateTime,
            default=datetime.utcnow,
            onupdate=datetime.utcnow,
            nullable=False,
        )
    )


class UUIDMixin(BaseModel):
    id: UUID4 = Field(
        default_factory=uuid4, primary_key=True, index=True, nullable=False
    )


class BaseNamespace(SQLModel):
    name: str = Field(..., min_length=1, max_length=255, unique=True)

    class Config:
        schema_extra = {
            "example": {
                "name": "hudson",
            }
        }


class NamespaceCreate(BaseNamespace):
    ...


class Namespace(BaseNamespace, UUIDMixin, TimestampsMixin, table=True):
    __tablename__ = "namespaces"

    topics: List["Topic"] = Relationship(
        back_populates="namespace",
        sa_relationship_kwargs={"lazy": "selectin", "cascade": "all, delete-orphan"},
    )
    datasets: List["Dataset"] = Relationship(
        back_populates="namespace",
        sa_relationship_kwargs={"lazy": "selectin", "cascade": "all, delete-orphan"},
    )


class NamespaceRead(BaseNamespace, UUIDMixin, TimestampsMixin):
    topics: List["Topic"]
    datasets: List["Dataset"]


class BaseTopic(SQLModel):
    name: str

    class Config:
        schema_extra = {
            "example": {
                "name": "default",
            }
        }


class TopicCreate(BaseTopic):
    namespace_id: UUID4


class Topic(BaseTopic, UUIDMixin, TimestampsMixin, table=True):
    __tablename__ = "topics"

    __table_args__ = (UniqueConstraint("namespace_id", "name"),)

    namespace_id: UUID4 = Field(
        default=None, foreign_key="namespaces.id", nullable=False
    )
    namespace: Namespace = Relationship(
        back_populates="topics", sa_relationship_kwargs={"lazy": "selectin"}
    )
    subscriptions: List["Subscription"] = Relationship(
        back_populates="topic",
        sa_relationship_kwargs={"lazy": "selectin", "cascade": "all, delete-orphan"},
    )


class TopicRead(BaseTopic, UUIDMixin, TimestampsMixin):
    namespace: Namespace
    subscriptions: List["Subscription"]


class BaseSubscription(SQLModel):
    name: str
    delivery_type: DeliveryType
    # TODO: needed for pull
    # message_retention_duration_minutes: int = Field(..., ge=0, le=7 * 24 * 60)
    push_endpoint: Optional[AnyHttpUrl]

    @validator("push_endpoint", pre=True, always=True)
    def validate_push_endpoint_https(cls, v: AnyHttpUrl) -> AnyHttpUrl:
        if v is not None and not v.startswith("https://"):
            raise ValueError("push_endpoint must be a HTTPS URL")
        return v

    class Config:
        schema_extra = {
            "example": {
                "name": "default",
                "delivery_type": "push",
            }
        }


class SubscriptionCreate(BaseSubscription):
    topic_id: UUID4


class Subscription(BaseSubscription, UUIDMixin, TimestampsMixin, table=True):
    __tablename__ = "subscriptions"

    __table_args__ = (UniqueConstraint("topic_id", "name"),)

    topic_id: UUID4 = Field(default=None, foreign_key="topics.id", nullable=False)
    topic: Topic = Relationship(
        back_populates="subscriptions", sa_relationship_kwargs={"lazy": "selectin"}
    )


class SubscriptionRead(BaseSubscription, UUIDMixin, TimestampsMixin):
    topic: Topic


class BaseDataset(SQLModel):
    name: str
    description: Optional[str] = Field(default=None, nullable=True)

    class Config:
        schema_extra = {
            "example": {
                "name": "default",
                "description": "default dataset",
            }
        }


class DatasetCreate(BaseDataset):
    namespace_id: UUID4


class Dataset(BaseDataset, UUIDMixin, TimestampsMixin, table=True):
    __tablename__ = "datasets"

    __table_args__ = (UniqueConstraint("namespace_id", "name"),)

    namespace_id: UUID4 = Field(
        default=None, foreign_key="namespaces.id", nullable=False
    )
    namespace: Namespace = Relationship(
        back_populates="datasets", sa_relationship_kwargs={"lazy": "selectin"}
    )


class DatasetRead(BaseDataset, UUIDMixin, TimestampsMixin):
    namespace: Namespace


NamespaceRead.update_forward_refs()
TopicRead.update_forward_refs()



import base64
import json
from uuid import uuid4

from httpx import AsyncClient


async def test_publish_message_push(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    topic = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions",
        json={
            "name": "default",
            "topic_id": topic["id"],
            "delivery_type": "push",
            "push_endpoint": "https://example.com/default",
        },
    )
    assert response.status_code == 200

    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions",
        json={
            "name": "default2",
            "topic_id": topic["id"],
            "delivery_type": "push",
            "push_endpoint": "https://example.com/default2",
        },
    )
    assert response.status_code == 200

    data = base64.b64encode(
        json.dumps({"message": "Hello world!"}).encode("utf-8")
    ).decode("utf-8")
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/publish",
        json={"data": data},
    )
    assert response.status_code == 200


async def test_publish_message_missing_namespace_fails(
    client: AsyncClient,
) -> None:
    response = await client.post(
        f"/namespaces/{uuid4()}/topics/{uuid4()}/publish",
        json={"data": "eyJtc2ciOiJoZWxsbyB3b3JsZCEifQo="},
    )
    assert response.status_code == 400


async def test_publish_missing_topic_fails(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{uuid4()}/publish",
        json={"data": "eyJtc2ciOiJoZWxsbyB3b3JsZCEifQo="},
    )
    assert response.status_code == 400


async def test_publish_message_too_big_fails(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    topic = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/publish",
        json={
            "data": base64.b64encode(("A" * 10_000_000).encode("utf8")).decode("utf8")
        },
    )
    assert response.status_code == 422
    assert response.json() == {
        "detail": [
            {
                "loc": ["body", "data"],
                "msg": "Message data must be less than or equal to 10MB.",
                "type": "value_error",
            }
        ]
    }


async def test_publish_message_invalid_b64_fails(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    topic = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/publish",
        json={"data": "eyJpbnZhbGlkIjogImpzbw=="},
    )
    assert response.status_code == 422
    assert response.json() == {
        "detail": [
            {
                "loc": ["body", "data"],
                "msg": "Message data must be a valid base64 encoded JSON string.",
                "type": "value_error",
            }
        ]
    }



from typing import Any, List, Optional, Union

from docarray import Document, DocumentArray
from httpx import Client
from httpx._types import HeaderTypes, QueryParamTypes, RequestContent, RequestData
from pydantic import UUID4

from hudson._env import env
from hudson._types import DataArray
from hudson.exc import BaseHudsonException
from hudson.models import DatasetRead, NamespaceRead


class HudsonClient(object):
    def __init__(self, url: str) -> None:
        self.url = url

    def request(
        self,
        method: str,
        path: str,
        content: Optional[RequestContent] = None,
        data: Optional[RequestData] = None,
        json: Optional[Any] = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
    ) -> Any:
        with Client() as c:
            response = c.request(
                method=method,
                url=self.url + path,
                content=content,
                data=data,
                json=json,
                params=params,
                headers=headers,
                timeout=env.TIMEOUT_SECONDS,
            )
            if not response.is_success:
                raise BaseHudsonException(response.text)
            return response.json()

    def create_namespace(self, name: str) -> NamespaceRead:
        response = self.request(method="POST", path="/namespaces", json={"name": name})
        return NamespaceRead(**response)

    def get_namespace(self, namespace_id: UUID4) -> NamespaceRead:
        response = self.request(method="GET", path=f"/namespaces/{namespace_id}")
        return NamespaceRead(**response)

    def delete_namespace(self, namespace_id: UUID4) -> NamespaceRead:
        response = self.request(method="DELETE", path=f"/namespaces/{namespace_id}")
        return NamespaceRead(**response)

    def list_namespaces(self, name: str) -> List[NamespaceRead]:
        response = self.request(method="GET", path="/namespaces", params={"name": name})
        return [NamespaceRead(**item) for item in response]

    def create_dataset(
        self, namespace_id: UUID4, name: str, description: Optional[str] = None
    ) -> DatasetRead:
        _json = {"name": name, "namespace_id": str(namespace_id)}
        if description is not None:
            _json["description"] = description

        response = self.request(
            method="POST",
            path=f"/namespaces/{namespace_id}/datasets",
            json=_json,
        )

        return DatasetRead(**response)

    def get_dataset(self, namespace_id: UUID4, dataset_id: UUID4) -> DatasetRead:
        response = self.request(
            method="GET", path=f"/namespaces/{namespace_id}/datasets/{dataset_id}"
        )
        return DatasetRead(**response)

    def delete_dataset(self, namespace_id: UUID4, dataset_id: UUID4) -> DatasetRead:
        response = self.request(
            method="DELETE", path=f"/namespaces/{namespace_id}/datasets/{dataset_id}"
        )
        return DatasetRead(**response)

    def write_dataset(
        self,
        namespace_id: UUID4,
        dataset_id: UUID4,
        data: DocumentArray,
    ) -> None:
        _data = [d.to_pydantic_model().dict() for d in data]
        self.request(
            "POST",
            f"/namespaces/{namespace_id}/datasets/{dataset_id}/write",
            json={"data": _data},
        )
        return None

    def read_dataset(
        self,
        namespace_id: UUID4,
        dataset_id: UUID4,
        as_document_array: bool = True,
    ) -> Union[DocumentArray, DataArray]:
        response = self.request(
            method="GET",
            path=f"/namespaces/{namespace_id}/datasets/{dataset_id}/read",
        )
        da = DataArray(**response)
        if as_document_array:
            return DocumentArray([Document.from_pydantic_model(d) for d in da.data])
        return da


hudson_client = HudsonClient(url=env.HUDSON_SERVER_URL)



from datetime import datetime
from typing import Any
from uuid import uuid4

from hudson.server.log import log


def test_logger(capsys: Any) -> None:
    log.exception("this is an exception")

    log.info(
        "testing types",
        extra={
            "string": "string",
            "bytes": b"test",
            "set": {"test", "test2"},
            "time": datetime.now(),
            "uuid": uuid4(),
        },
    )

    class MyUnsupportedType:
        def __init__(self, data: str) -> None:
            self.data = data

    log.info("testing types", extra={"random": MyUnsupportedType("test")})
    out, err = capsys.readouterr()
    assert out == ""
    assert "TypeError: Object of type MyUnsupportedType is not JSON serializable" in err



from uuid import uuid4

from httpx import AsyncClient
from sqlalchemy import text
from sqlmodel.ext.asyncio.session import AsyncSession


async def test_create_topics(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    namespace_id = response.json()["id"]
    response = await client.post(
        f"/namespaces/{namespace_id}/topics",
        json={"name": "test", "namespace_id": namespace_id},
    )
    assert response.status_code == 200
    assert response.json()["name"] == "test"
    assert response.json()["namespace"]["id"] == namespace_id
    assert response.json()["id"] is not None
    assert response.json()["created_at"] is not None


async def test_get_topics(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    namespace_id = response.json()["id"]
    response = await client.post(
        f"/namespaces/{namespace_id}/topics",
        json={"name": "test", "namespace_id": namespace_id},
    )
    assert response.status_code == 200
    topic_id = response.json()["id"]
    response = await client.get(f"/namespaces/{namespace_id}/topics")
    assert response.status_code == 200
    assert response.json()[0]["id"] == topic_id
    assert response.json()[0]["name"] == "test"
    assert response.json()[0]["namespace"]["id"] == namespace_id


async def test_delete_topics(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    namespace_id = response.json()["id"]
    response = await client.post(
        f"/namespaces/{namespace_id}/topics",
        json={"name": "test", "namespace_id": namespace_id},
    )
    assert response.status_code == 200
    topic_id = response.json()["id"]
    response = await client.delete(f"/namespaces/{namespace_id}/topics/{topic_id}")
    assert response.status_code == 200
    assert response.json()["id"] == topic_id
    assert response.json()["name"] == "test"
    assert response.json()["namespace"]["id"] == namespace_id
    response = await client.get(f"/namespaces/{namespace_id}/topics")
    assert response.status_code == 200
    assert len(response.json()) == 0


async def test_delete_namespace_cascades_topics(
    client: AsyncClient, async_db_session: AsyncSession
) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    namespace_id = response.json()["id"]
    response = await client.post(
        f"/namespaces/{namespace_id}/topics",
        json={"name": "test", "namespace_id": namespace_id},
    )
    assert response.status_code == 200

    async with async_db_session.begin():
        result = await async_db_session.execute(text("SELECT * FROM topics"))
        assert len(result.scalars().all()) == 1

    response = await client.delete(f"/namespaces/{namespace_id}")
    assert response.status_code == 200
    assert response.json()["id"] == namespace_id
    assert response.json()["name"] == "test"

    async with async_db_session.begin():
        result = await async_db_session.execute(text("SELECT * FROM topics"))
        assert result.scalars().all() == []


async def test_get_topic_with_query(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    namespace_id = response.json()["id"]
    response = await client.post(
        f"/namespaces/{namespace_id}/topics",
        json={"name": "test", "namespace_id": namespace_id},
    )
    assert response.status_code == 200
    topic_id = response.json()["id"]
    response = await client.get(f"/namespaces/{namespace_id}/topics?name=tes")
    assert response.status_code == 200
    assert response.json()[0]["id"] == topic_id
    assert response.json()[0]["name"] == "test"
    assert response.json()[0]["namespace"]["id"] == namespace_id


async def test_create_topic_already_exists(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    namespace_id = response.json()["id"]

    response = await client.post(
        f"/namespaces/{namespace_id}/topics",
        json={"name": "test", "namespace_id": namespace_id},
    )
    assert response.status_code == 200
    topic_id = response.json()["id"]

    response = await client.post(
        f"/namespaces/{namespace_id}/topics",
        json={"name": "test", "namespace_id": namespace_id},
    )
    assert response.status_code == 200
    assert response.json()["id"] == topic_id

    # assert there is one topic
    response = await client.get(f"/namespaces/{namespace_id}/topics")
    assert response.status_code == 200
    assert len(response.json()) == 1


async def test_get_topics_missing_namespace_fails(client: AsyncClient) -> None:
    response = await client.get(f"/namespaces/{uuid4()}/topics")
    assert response.status_code == 400
    assert response.json()["detail"] == "Namespace not found."


async def test_create_topic_missing_namespace_fails(client: AsyncClient) -> None:
    ns_id = str(uuid4())
    response = await client.post(
        f"/namespaces/{ns_id}/topics", json={"name": "test", "namespace_id": ns_id}
    )
    assert response.status_code == 400
    assert response.json()["detail"] == "Namespace not found."


async def test_delete_topic_missing_namespace_fails(client: AsyncClient) -> None:
    response = await client.delete(f"/namespaces/{uuid4()}/topics/{uuid4()}")
    assert response.status_code == 400
    assert response.json()["detail"] == "Namespace not found."


async def test_delete_topic_not_found_fails(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    namespace_id = response.json()["id"]
    response = await client.delete(f"/namespaces/{namespace_id}/topics/{uuid4()}")
    assert response.status_code == 400
    assert response.json()["detail"] == "Topic not found."



import asyncio
import datetime
import json
import logging
import socket
from logging.handlers import QueueHandler
from queue import SimpleQueue
from typing import Any, List
from uuid import UUID

from hudson._env import env


class LogEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, bytes):
            return obj.decode("utf-8")
        elif isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, (set, frozenset)):
            return tuple(obj)
        elif isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
            return obj.isoformat()
        return super().default(obj)


class StructuredMessage:
    def __init__(self, **kwargs: Any) -> None:
        self.kwargs = kwargs

    def __str__(self) -> str:
        return LogEncoder().encode(self.kwargs)


class OnelineFormatter(logging.Formatter):
    def formatException(self, exc_info: Any) -> str:
        result = super().formatException(exc_info)
        return repr(result)

    def format(self, record: logging.LogRecord) -> str:
        result = super().format(record)
        if record.exc_text:
            result = result.replace("\n", "")
        result_dict = record.__dict__
        result_dict["host"] = socket.gethostname()
        return str(StructuredMessage(**result_dict))


class StructuredLogger:
    DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

    @staticmethod
    def create_logger() -> logging.Logger:
        logger = logging.getLogger(__name__)
        log_handler = logging.StreamHandler()
        formatter = OnelineFormatter(datefmt=StructuredLogger.DATE_FORMAT)
        log_handler.setFormatter(formatter)
        logger.addHandler(log_handler)
        logger.setLevel(env.LOG_LEVEL.upper())
        return logger


class LocalQueueHandler(QueueHandler):
    def emit(self, record: logging.LogRecord) -> None:  # pragma: no cover
        try:
            self.enqueue(record)
        except asyncio.CancelledError:
            raise
        except Exception:
            self.handleError(record)


def setup_logging_queue() -> None:
    """Move log handlers to a separate thread.

    Replace handlers on the root logger with a LocalQueueHandler,
    and start a logging.QueueListener holding the original
    handlers.

    https://www.zopatista.com/python/2019/05/11/asyncio-logging/
    """
    root = logging.getLogger()
    queue: SimpleQueue = SimpleQueue()
    handlers: List[logging.Handler] = []
    handler = LocalQueueHandler(queue)

    root.addHandler(handler)
    for h in root.handlers[:]:
        if h is not handler:  # pragma: no cover
            root.removeHandler(h)
            handlers.append(h)

    listener = logging.handlers.QueueListener(
        queue, *handlers, respect_handler_level=True
    )
    listener.start()


log = StructuredLogger.create_logger()
setup_logging_queue()



import os
from uuid import uuid4

from httpx import AsyncClient

from hudson._env import env


async def test_create_namespace(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    assert response.json()["name"] == "test"


async def test_create_namespace_no_duplicates(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    assert response.json()["name"] == "test"
    ns_id = response.json()["id"]
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    assert response.json()["name"] == "test"
    assert response.json()["id"] == ns_id
    response = await client.get("/namespaces")
    assert response.status_code == 200
    assert len(response.json()) == 1


async def test_get_namespaces(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    assert response.json()["name"] == "test"
    response = await client.get("/namespaces")
    assert response.status_code == 200
    assert len(response.json()) == 1
    assert response.json()[0]["name"] == "test"


async def test_get_namespace(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    assert response.json()["name"] == "test"
    ns_id = response.json()["id"]
    response = await client.get(f"/namespaces/{ns_id}")
    assert response.status_code == 200
    assert response.json()["name"] == "test"


async def test_get_namespace_not_found(client: AsyncClient) -> None:
    response = await client.get(f"/namespaces/{uuid4()}")
    assert response.status_code == 400


async def test_get_namespaces_q(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "test"})
    assert response.status_code == 200
    assert response.json()["name"] == "test"
    ns_id = response.json()["id"]
    response = await client.get("/namespaces?name=test")
    assert response.status_code == 200
    assert len(response.json()) == 1
    assert response.json()[0]["name"] == "test"
    assert ns_id == response.json()[0]["id"]


async def test_get_namespaces_q_no_match(client: AsyncClient) -> None:
    response = await client.get("/namespaces?name=foo")
    assert response.status_code == 200
    assert len(response.json()) == 0


async def test_get_namespaces_q_multiple(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "foo"})
    assert response.status_code == 200
    assert response.json()["name"] == "foo"
    response = await client.get("/namespaces?name=foo")
    assert response.status_code == 200
    assert len(response.json()) == 1
    assert response.json()[0]["name"] == "foo"


async def test_get_namespaces_q_multiple_no_match(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "foo"})
    assert response.status_code == 200
    assert response.json()["name"] == "foo"
    response = await client.get("/namespaces?name=bar")
    assert response.status_code == 200
    assert len(response.json()) == 0


async def test_delete_namespaces(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "foo"})
    assert response.status_code == 200
    namespace_id = response.json()["id"]

    # create a dataset in the namespace
    response = await client.post(
        f"/namespaces/{namespace_id}/datasets",
        json={"name": "foo", "namespace_id": str(namespace_id)},
    )
    assert response.status_code == 200

    response = await client.delete(f"/namespaces/{namespace_id}")
    assert response.status_code == 200

    response = await client.get("/namespaces")
    assert response.status_code == 200
    assert len(response.json()) == 0

    # assert that the namespace directory is gone
    assert not os.path.exists(f"{env.DATASETS_PATH}/{namespace_id}")


async def test_delete_namespace_not_found(client: AsyncClient) -> None:
    response = await client.delete(f"/namespaces/{uuid4()}")
    assert response.status_code == 400
    assert response.status_code == 400



import typer
import uvicorn

from hudson import __version__
from hudson.const import APP_IMPORT_STRING, HUDSON

name = f"{HUDSON} {__version__}"

app = typer.Typer(
    name=name,
    rich_markup_mode="markdown",
    no_args_is_help=True,
)


@app.callback()
def main_callback() -> None:
    pass


main_callback.__doc__ = name


@app.command("version")
def _version() -> None:
    """Print the version of Hudson."""
    typer.echo(__version__, nl=False)


@app.command("server")
def _server(
    host: str = typer.Option(
        "127.0.0.1",
        "--host",
        help="Bind socket to this host.",
    ),
    port: int = typer.Option(
        8000,
        "--port",
        help="Bind socket to this port.",
    ),
    reload: bool = typer.Option(
        False,
        help="Reload the server on code changes.",
    ),
    workers: int = typer.Option(
        1,
        "--workers",
        help="Number of workers to run.",
    ),
) -> None:
    """Start the Hudson server."""
    uvicorn.run(  # pragma: no cover
        app=APP_IMPORT_STRING,
        port=port,
        host=host,
        reload=reload,
        workers=workers,
    )


@app.command("deploy")
def _deploy(
    name: str = typer.Argument(..., help="Name of the project to deploy."),
    stub_filepath: str = typer.Option(
        "hudson._modal",
        "--stub-path",
        "-s",
        help="Filepath reference to the stub to deploy. "
        "Relative the project module root.",
    ),
) -> None:  # pragma: no cover
    """Deploy Hudson to Modal."""
    from modal.cli.app import deploy

    deploy(
        name=name,
        stub_ref=stub_filepath,
    )



from typing import Any, Optional
from uuid import uuid4

from hudson._types import DataArray


class MockResponse:
    def __init__(
        self,
        status_code: int,
        json: Any = None,
        is_success: bool = True,
        text: Optional[str] = None,
    ) -> None:
        self.status_code = status_code
        self._json = json
        self.is_success = not str(status_code).startswith(("4", "5"))
        self.text = text

    def json(self) -> Any:
        return self._json


MockNamespace = {
    "id": uuid4(),
    "name": "default",
    "created_at": "2021-01-01T00:00:00+00:00",
    "updated_at": "2021-01-01T00:00:00+00:00",
    "topics": [],
    "datasets": [],
}


MockDataset = {
    "id": uuid4(),
    "name": "default",
    "namespace_id": uuid4(),
    "created_at": "2021-01-01T00:00:00+00:00",
    "updated_at": "2021-01-01T00:00:00+00:00",
    "namespace": MockNamespace,
}


MockDataArray = DataArray.Config.schema_extra["example"]



from unittest import mock
from uuid import UUID

import pytest
from docarray import Document, DocumentArray

from hudson import hudson_client
from hudson._types import DataArray
from hudson.exc import BaseHudsonException
from tests.mocks import MockDataArray, MockDataset, MockNamespace, MockResponse


def test_base_exception() -> None:
    assert issubclass(BaseHudsonException, Exception)


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=200, json=MockNamespace),
)
async def test_create_namespace(mock_request: mock.MagicMock) -> None:
    ns = hudson_client.create_namespace("default")
    assert ns.name == "default"


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=200, json=MockNamespace),
)
async def test_get_namespace(mock_request: mock.MagicMock) -> None:
    ns = hudson_client.get_namespace(UUID(str(MockNamespace["id"])))
    assert ns.name == "default"


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=200, json=MockNamespace),
)
async def test_delete_namespace(mock_request: mock.MagicMock) -> None:
    ns = hudson_client.delete_namespace(UUID(str(MockNamespace["id"])))
    assert ns.name == "default"


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=200, json=[MockNamespace]),
)
async def test_list_namespaces(mock_request: mock.MagicMock) -> None:
    ns = hudson_client.list_namespaces("default")
    assert ns[0].name == "default"


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=200, json=MockDataset),
)
async def test_create_dataset(mock_request: mock.MagicMock) -> None:
    ds = hudson_client.create_dataset(UUID(str(MockNamespace["id"])), "default")
    assert ds.name == "default"


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=200, json=MockDataset),
)
async def test_create_dataset_with_description(mock_request: mock.MagicMock) -> None:
    _ds = MockDataset.copy()
    _ds["description"] = "test"
    mock_request.return_value._json = _ds
    ds = hudson_client.create_dataset(UUID(str(MockNamespace["id"])), "default", "test")
    assert ds.name == "default"
    assert ds.description == "test"


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=200, json=MockDataset),
)
async def test_get_dataset(mock_request: mock.MagicMock) -> None:
    ds = hudson_client.get_dataset(
        namespace_id=UUID(str(MockNamespace["id"])),
        dataset_id=UUID(str(MockDataset["id"])),
    )
    assert ds.name == "default"


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=200, json=MockDataset),
)
async def test_delete_dataset(mock_request: mock.MagicMock) -> None:
    ds = hudson_client.delete_dataset(
        namespace_id=UUID(str(MockNamespace["id"])),
        dataset_id=UUID(str(MockDataset["id"])),
    )
    assert ds.name == "default"


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=200),
)
async def test_write_dataset(mock_request: mock.MagicMock) -> None:
    hudson_client.write_dataset(
        namespace_id=UUID(str(MockNamespace["id"])),
        dataset_id=UUID(str(MockDataset["id"])),
        data=DocumentArray([Document(text="hello world")]),
    )


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=500, text="Internal Server Error"),
)
async def test_client_returns_error(mock_request: mock.MagicMock) -> None:
    with pytest.raises(BaseHudsonException):
        hudson_client.write_dataset(
            namespace_id=UUID(str(MockNamespace["id"])),
            dataset_id=UUID(str(MockDataset["id"])),
            data=DocumentArray([Document(text="hello world")]),
        )


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=200, json=MockDataArray),
)
async def test_read_dataset(mock_request: mock.MagicMock) -> None:
    ds = hudson_client.read_dataset(
        namespace_id=UUID(str(MockNamespace["id"])),
        dataset_id=UUID(str(MockDataset["id"])),
    )
    assert len(ds) == 2  # type: ignore


@mock.patch(
    "hudson.client.main.Client.request",
    return_value=MockResponse(status_code=200, json=MockDataArray),
)
async def test_read_dataset_not_as_document_arr(mock_request: mock.MagicMock) -> None:
    ds = hudson_client.read_dataset(
        namespace_id=UUID(str(MockNamespace["id"])),
        dataset_id=UUID(str(MockDataset["id"])),
        as_document_array=False,
    )
    assert type(ds) == DataArray
    assert len(ds.data) == 2



import json
from typing import Callable

from fastapi import Request, Response
from fastapi.routing import APIRoute

from hudson._types import RequestLoggerMessage, ResponseLoggerMessage
from hudson.server.log import log


class _APIRoute(APIRoute):
    """_APIRoute.

    _APIRoute is a custom APIRoute class that adds a background task to the
    response to log request and response data.
    """

    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def _log(req: RequestLoggerMessage, res: ResponseLoggerMessage) -> None:
            log.debug({"req": json.loads(req.json())})
            log.debug({"res": json.loads(res.json())})

        async def custom_route_handler(request: Request) -> Response:
            req = RequestLoggerMessage(**request.__dict__)
            response = await original_route_handler(request)
            res = ResponseLoggerMessage(**response.__dict__)
            await _log(req=req, res=res)
            return response

        return custom_route_handler



from hudson._env import _Env


def test_env_conversion() -> None:
    test_env = _Env(
        API_SECRET_KEY="test",
        PSQL_URL="postgresql+asyncpg://hudson:hudson@localhost:5432/hudson",
        PSQL_POOL_SIZE=5,
        PSQL_MAX_OVERFLOW=10,
        PSQL_POOL_PRE_PING=True,
    )
    s = test_env.to_modal_secret()
    assert s["API_SECRET_KEY"] == "test"
    assert s["PSQL_URL"] == "postgresql+asyncpg://hudson:hudson@localhost:5432/hudson"
    assert s["PSQL_POOL_SIZE"] == "5"
    assert s["PSQL_MAX_OVERFLOW"] == "10"
    assert s["PSQL_POOL_PRE_PING"] == "true"



import os

from fastapi import FastAPI

from hudson import __version__
from hudson.const import HUDSON
from hudson.server import routers

os.environ["TZ"] = "UTC"

app = FastAPI(title=HUDSON, version=__version__)

app.include_router(routers.health_router)
app.include_router(routers.namespace_router)
app.include_router(routers.pubsub_router)
app.include_router(routers.dataset_router)



from uuid import uuid4

from httpx import AsyncClient
from sqlmodel import select
from sqlmodel.ext.asyncio.session import AsyncSession

from hudson.models import Subscription


async def test_create_subscription(
    client: AsyncClient, async_db_session: AsyncSession
) -> None:
    response = await client.post("/namespaces", json={"name": "hudson"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    topic = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions",
        json={
            "name": "default",
            "topic_id": topic["id"],
            "delivery_type": "push",
            "push_endpoint": "https://localhost:4242",
        },
    )
    assert response.status_code == 200
    subscription = response.json()
    assert subscription["name"] == "default"
    assert subscription["topic"]["id"] == topic["id"]
    assert subscription["topic"]["namespace_id"] == namespace["id"]

    async with async_db_session.begin():
        result = await async_db_session.execute(select(Subscription))
        subs = result.scalars().all()
        assert len(subs) == 1
        assert str(subs[0].topic_id) == topic["id"]


async def test_create_subscription_that_already_exists(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "hudson"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    topic = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions",
        json={
            "name": "default",
            "topic_id": topic["id"],
            "delivery_type": "push",
            "push_endpoint": "https://localhost:4242",
        },
    )
    assert response.status_code == 200
    subscription = response.json()
    assert subscription["name"] == "default"
    assert subscription["topic"]["id"] == topic["id"]
    assert subscription["topic"]["namespace_id"] == namespace["id"]

    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions",
        json={
            "name": "default",
            "topic_id": topic["id"],
            "delivery_type": "push",
            "push_endpoint": "https://localhost:4242",
        },
    )
    assert response.status_code == 200
    assert response.json() == subscription


async def test_get_subscriptions(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "hudson"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    topic = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions",
        json={
            "name": "default",
            "topic_id": topic["id"],
            "delivery_type": "push",
            "push_endpoint": "https://localhost:4242",
        },
    )
    assert response.status_code == 200
    subscription = response.json()
    response = await client.get(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions"
    )
    assert response.status_code == 200
    subscriptions = response.json()
    assert len(subscriptions) == 1
    assert subscriptions[0]["name"] == subscription["name"]
    assert subscriptions[0]["topic"]["id"] == subscription["topic"]["id"]
    assert (
        subscriptions[0]["topic"]["namespace_id"]
        == subscription["topic"]["namespace_id"]
    )

    response = await client.get(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions?name=default"
    )
    assert response.status_code == 200
    assert response.json() == subscriptions


async def test_delete_subscription(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "hudson"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    topic = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions",
        json={
            "name": "default",
            "topic_id": topic["id"],
            "delivery_type": "push",
            "push_endpoint": "https://localhost:4242",
        },
    )
    assert response.status_code == 200
    subscription = response.json()
    response = await client.delete(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions/{subscription['id']}"
    )
    assert response.status_code == 200
    response = await client.get(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions"
    )
    assert response.status_code == 200
    subscriptions = response.json()
    assert len(subscriptions) == 0


async def test_create_subscription_missing_namespace_fails(client: AsyncClient) -> None:
    response = await client.post(
        f"/namespaces/{str(uuid4())}/topics/{str(uuid4())}/subscriptions",
        json={
            "name": "default",
            "topic_id": str(uuid4()),
            "delivery_type": "push",
            "push_endpoint": "https://localhost:4242",
        },
    )
    assert response.status_code == 400
    assert response.json()["detail"] == "Namespace not found."


async def test_create_subscription_missing_topic_fails(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "hudson"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{str(uuid4())}/subscriptions",
        json={
            "name": "default",
            "topic_id": str(uuid4()),
            "delivery_type": "push",
            "push_endpoint": "https://localhost:4242",
        },
    )
    assert response.status_code == 400
    assert response.json()["detail"] == "Topic not found."


async def test_get_subscriptions_missing_namespace_fails(client: AsyncClient) -> None:
    response = await client.get(
        f"/namespaces/{str(uuid4())}/topics/{str(uuid4())}/subscriptions"
    )
    assert response.status_code == 400
    assert response.json()["detail"] == "Namespace not found."


async def test_get_subscriptions_missing_topic_fails(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "hudson"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.get(
        f"/namespaces/{namespace['id']}/topics/{str(uuid4())}/subscriptions"
    )
    assert response.status_code == 400
    assert response.json()["detail"] == "Topic not found."


async def test_delete_subscription_missing_namespace_fails(client: AsyncClient) -> None:
    response = await client.delete(
        f"/namespaces/{str(uuid4())}/topics/{str(uuid4())}/subscriptions/{str(uuid4())}"
    )
    assert response.status_code == 400
    assert response.json()["detail"] == "Namespace not found."


async def test_delete_subscription_missing_topic_fails(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "hudson"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.delete(
        f"/namespaces/{namespace['id']}/topics/{str(uuid4())}/subscriptions/{str(uuid4())}"
    )
    assert response.status_code == 400
    assert response.json()["detail"] == "Topic not found."


async def test_delete_subscription_missing_subscription_fails(
    client: AsyncClient,
) -> None:
    response = await client.post("/namespaces", json={"name": "hudson"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    topic = response.json()
    response = await client.delete(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions/{str(uuid4())}"
    )
    assert response.status_code == 400
    assert response.json()["detail"] == "Subscription not found."


async def test_delete_namespace_cascades_to_subscriptions(
    client: AsyncClient, async_db_session: AsyncSession
) -> None:
    response = await client.post("/namespaces", json={"name": "hudson"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    topic = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions",
        json={
            "name": "default",
            "topic_id": topic["id"],
            "delivery_type": "push",
            "push_endpoint": "https://localhost:4242",
        },
    )
    assert response.status_code == 200
    subscription = response.json()
    response = await client.delete(f"/namespaces/{namespace['id']}")
    assert response.status_code == 200

    async with async_db_session as session:
        assert (
            await session.execute(
                select(Subscription).where(Subscription.id == subscription["id"])
            )
        ).scalar_one_or_none() is None


async def test_create_subscription_needs_https_endpoint(
    client: AsyncClient, async_db_session: AsyncSession
) -> None:
    response = await client.post("/namespaces", json={"name": "hudson"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    topic = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/topics/{topic['id']}/subscriptions",
        json={
            "name": "default",
            "topic_id": topic["id"],
            "delivery_type": "push",
            "push_endpoint": "http://localhost:4242",
        },
    )
    assert response.status_code == 422
    assert response.json()["detail"][0]["msg"] == "push_endpoint must be a HTTPS URL"



import os
from uuid import uuid4

from docarray import Document, DocumentArray
from httpx import AsyncClient

from hudson._env import env


async def test_create_dataset(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    dataset = response.json()
    assert dataset["name"] == "default"
    assert dataset["namespace"]["id"] == namespace["id"]


async def test_create_dataset_with_description(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets",
        json={
            "name": "default",
            "namespace_id": namespace["id"],
            "description": "test",
        },
    )
    assert response.status_code == 200
    dataset = response.json()
    assert dataset["name"] == "default"
    assert dataset["namespace"]["id"] == namespace["id"]
    assert dataset["description"] == "test"


async def test_get_datasets(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    response = await client.get(f"/namespaces/{namespace['id']}/datasets")
    assert response.status_code == 200
    assert len(response.json()) == 1


async def test_get_dataset(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    dataset = response.json()
    response = await client.get(
        f"/namespaces/{namespace['id']}/datasets/{dataset['id']}"
    )
    assert response.status_code == 200
    assert response.json()["name"] == "default"
    assert response.json()["namespace"]["id"] == namespace["id"]


async def test_get_missing_dataset_in_namespace_fails(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.get(f"/namespaces/{namespace['id']}/datasets/{uuid4()}")
    assert response.status_code == 400


async def test_missing_dataset_fails(client: AsyncClient) -> None:
    response = await client.get(f"/namespaces/{uuid4()}/datasets/{uuid4()}")
    assert response.status_code == 400


async def test_get_dataset_name_query(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    dataset = response.json()
    response = await client.get(
        f"/namespaces/{namespace['id']}/datasets?name={dataset['name']}"
    )
    assert response.status_code == 200
    dataset = response.json()[0]
    assert dataset["name"] == "default"
    assert dataset["namespace"]["id"] == namespace["id"]


async def test_create_dataset_missing_namespace_fails(client: AsyncClient) -> None:
    ns_id = str(uuid4())
    response = await client.post(
        f"/namespaces/{ns_id}/datasets",
        json={"name": "default", "namespace_id": ns_id},
    )
    assert response.status_code == 400


async def test_create_dataset_missing_name_fails(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets",
        json={"namespace_id": namespace["id"]},
    )
    assert response.status_code == 422


async def test_no_duplicate_datasets(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    response = await client.get(f"/namespaces/{namespace['id']}/datasets")
    assert response.status_code == 200
    datasets = response.json()
    assert len(datasets) == 1


async def test_delete_dataset(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    dataset = response.json()
    response = await client.delete(
        f"/namespaces/{namespace['id']}/datasets/{dataset['id']}"
    )
    assert response.status_code == 200
    response = await client.get(f"/namespaces/{namespace['id']}/datasets")
    assert response.status_code == 200
    datasets = response.json()
    assert len(datasets) == 0
    assert not os.path.exists(f"{env.DATASETS_PATH}/{namespace['id']}/{dataset['id']}")


async def test_get_dataset_missing_namespace_fails(client: AsyncClient) -> None:
    ns_id = str(uuid4())
    response = await client.get(f"/namespaces/{ns_id}/datasets")
    assert response.status_code == 400


async def test_delete_dataset_missing_namespace_fails(client: AsyncClient) -> None:
    ns_id = str(uuid4())
    ds_id = str(uuid4())
    response = await client.delete(f"/namespaces/{ns_id}/datasets/{ds_id}")
    assert response.status_code == 400


async def test_delete_missing_dataset_from_namespace_fails(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    ds_id = str(uuid4())
    response = await client.delete(f"/namespaces/{namespace['id']}/datasets/{ds_id}")
    assert response.status_code == 400


async def test_write_to_dataset(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    dataset = response.json()
    da = DocumentArray()
    da.extend([Document(text="hello"), Document(text="world!")])
    _json = {"data": [d.dict() for d in da.to_pydantic_model()]}
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets/{dataset['id']}/write",
        json=_json,
    )
    assert response.status_code == 200
    # count the number of files in the dataset path for this dataset
    dataset_path = f"{env.DATASETS_PATH}/{namespace['id']}/{dataset['id']}"
    assert len(os.listdir(dataset_path)) == 1


async def test_read_from_dataset(client: AsyncClient) -> None:
    response = await client.post("/namespaces", json={"name": "default"})
    assert response.status_code == 200
    namespace = response.json()
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets",
        json={"name": "default", "namespace_id": namespace["id"]},
    )
    assert response.status_code == 200
    dataset = response.json()
    da = DocumentArray()
    da.extend([Document(text="hello"), Document(text="world!")])
    _json = {"data": [d.dict() for d in da.to_pydantic_model()]}
    response = await client.post(
        f"/namespaces/{namespace['id']}/datasets/{dataset['id']}/write",
        json=_json,
    )
    assert response.status_code == 200
    response = await client.get(
        f"/namespaces/{namespace['id']}/datasets/{dataset['id']}/read"
    )
    assert response.status_code == 200
    assert response.json() == _json



import os

from httpx import AsyncClient

from hudson import __version__


async def test_tz() -> None:
    assert os.environ["TZ"] == "UTC"


async def test_healthcheck(client: AsyncClient) -> None:
    response = await client.get("/healthcheck")
    assert response.status_code == 200
    assert response.json()["message"] == "⛵️"
    assert response.json()["version"] == __version__



import modal
from fastapi import FastAPI

from hudson._env import _Env
from hudson.server.main import app

stub = modal.Stub()

_s = _Env(_env_file=".env")
stub["env"] = modal.Secret(_s.to_modal_secret())
image = modal.Image.debian_slim().pip_install_from_pyproject("pyproject.toml")


@stub.asgi(
    image=image,
    secret=stub["env"],
    shared_volumes={
        _s.DATASETS_PATH: modal.SharedVolume().persist(_s.MODAL_VOLUME_NAME)
    },
)
def _app() -> FastAPI:
    return app


if __name__ == "__main__":
    stub.serve()



from typing import AsyncGenerator

from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.orm import sessionmaker
from sqlmodel.ext.asyncio.session import AsyncSession

from hudson._env import env

async_psql_engine = create_async_engine(
    url=env.PSQL_URL,
    pool_size=env.PSQL_POOL_SIZE,
    max_overflow=env.PSQL_MAX_OVERFLOW,
    pool_pre_ping=env.PSQL_POOL_PRE_PING,
    future=True,
)


async def psql_db() -> AsyncGenerator:
    async_session = sessionmaker(
        async_psql_engine, class_=AsyncSession, expire_on_commit=False
    )
    async with async_session() as session:
        yield session



import os
import sys
from typing import Dict

from pydantic import BaseSettings, Field


class _Env(BaseSettings):
    LOG_LEVEL: str = Field(
        "INFO",
        env="LOG_LEVEL",
        description="Log level.",
    )
    HUDSON_SERVER_URL: str = Field(
        "http://localhost:8000",
        env="HUDSON_SERVER_URL",
        description="The Hudson server URL.",
    )
    API_SECRET_KEY: str = Field(
        ...,
        env="API_SECRET_KEY",
        description="The API secret key.",
    )
    PSQL_URL: str = Field(
        "postgresql+asyncpg://hudson:hudson@localhost:5432/hudson",
        env="PSQL_URL",
        description="The PSQL database URL.",
    )
    PSQL_POOL_SIZE: int = Field(
        5,
        env="PSQL_POOL_SIZE",
        description="The PSQL database pool size.",
    )
    PSQL_MAX_OVERFLOW: int = Field(
        10,
        env="PSQL_MAX_OVERFLOW",
        description="The PSQL database max overflow.",
    )
    PSQL_POOL_PRE_PING: bool = Field(
        True,
        env="PSQL_POOL_PRE_PING",
        description="The PSQL database pre pool ping.",
    )
    DATASETS_PATH: str = Field(
        f"{os.environ['HOME']}/.hudson/datasets",
        env="DATASETS_PATH",
        description="The root dir for storing datasets.",
    )
    TIMEOUT_SECONDS: int = Field(
        60,
        env="TIMEOUT_SECONDS",
        description="The timeout for HTTP requests.",
    )
    MODAL_VOLUME_NAME: str = Field(
        "hudson-datasets",
        env="MODAL_VOLUME_NAME",
        description="The Modal volume name for datasets.",
    )

    def to_modal_secret(self) -> Dict[str, str]:
        _result = self.dict()
        for k, v in _result.items():
            if isinstance(v, bool):
                _result[k] = str(v).lower()
            elif not isinstance(v, str):
                _result[k] = str(v)
        return _result

    class Config:
        env_file = ".env", ".env.local"
        env_encoding = "utf-8"


env = _Env()
if "pytest" in "".join(sys.argv):
    env = _Env(_env_file=".env.test")



from datetime import datetime
from typing import List, Optional

from fastapi import APIRouter, Body, Depends, HTTPException
from pydantic import UUID4
from sqlmodel.ext.asyncio.session import AsyncSession

from hudson import __version__
from hudson._types import DataArray, HealthResponse, Message
from hudson.db import psql_db
from hudson.models import (
    Dataset,
    DatasetCreate,
    DatasetRead,
    Namespace,
    NamespaceCreate,
    NamespaceRead,
    Subscription,
    SubscriptionCreate,
    SubscriptionRead,
    Topic,
    TopicCreate,
    TopicRead,
)
from hudson.server.log import log
from hudson.server.services import (
    datasets_service,
    namespace_service,
    subscriptions_service,
    topics_service,
)
from hudson.server.utils import _APIRoute

health_router = APIRouter(route_class=_APIRoute, tags=["health"])
dataset_router = APIRouter(route_class=_APIRoute, tags=["dataset"])
namespace_router = APIRouter(route_class=_APIRoute, tags=["namespace"])
pubsub_router = APIRouter(route_class=_APIRoute, tags=["pubsub"])


@health_router.get("/healthcheck", response_model=HealthResponse)
async def get_health() -> HealthResponse:
    """Get the health of the server.

    Returns:
        HealthResponse: The health of the server.
    """
    log.info("Healthcheck")
    return HealthResponse(message="⛵️", version=__version__, time=datetime.utcnow())


@namespace_router.post("/namespaces", response_model=NamespaceRead)
async def create_namespaces(
    namespace_create: NamespaceCreate = Body(...),
    psql: AsyncSession = Depends(psql_db),
) -> Namespace:
    """Create a namespace if it does not exist.

    Args:
        namespace_create (NamespaceCreate): The namespace to create.

    Returns:
        Namespace: The created namespace.
    """
    namespace = await namespace_service.get_by_name(
        name=namespace_create.name, psql=psql
    )
    if namespace is not None:
        return namespace
    return await namespace_service.create(namespace_create=namespace_create, psql=psql)


@namespace_router.get("/namespaces", response_model=List[NamespaceRead])
async def get_namespaces(
    name: Optional[str] = None, psql: AsyncSession = Depends(psql_db)
) -> List[Namespace]:
    """Get all namespaces.

    Args:
        name (Optional[str], optional): The namespace name. Defaults to None.

    Returns:
        List[Namespace]: The namespaces.
    """
    return await namespace_service.list(name=name, psql=psql)


@namespace_router.get("/namespaces/{namespace_id}", response_model=NamespaceRead)
async def get_namespace(
    namespace_id: UUID4, psql: AsyncSession = Depends(psql_db)
) -> Namespace:
    """Get a namespace.

    Args:
        namespace_id (UUID4): The namespace id.

    Returns:
        Namespace: The namespace.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    return namespace


@namespace_router.delete("/namespaces/{namespace_id}", response_model=NamespaceRead)
async def delete_namespaces(
    namespace_id: UUID4, psql: AsyncSession = Depends(psql_db)
) -> Namespace:
    """Delete a namespace.

    Args:
        namespace_id (UUID4): The namespace id.

    Returns:
        Namespace: The deleted namespace.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    return await namespace_service.delete(namespace_id=namespace_id, psql=psql)


@pubsub_router.get("/namespaces/{namespace_id}/topics", response_model=List[TopicRead])
async def get_topics(
    namespace_id: UUID4,
    name: Optional[str] = None,
    psql: AsyncSession = Depends(psql_db),
) -> List[Topic]:
    """Get all topics in a namespace.

    Args:
        namespace_id (UUID4): The namespace id.
        name (Optional[str], optional): The topic name. Defaults to None.

    Returns:
        List[Topic]: The topics.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    return await topics_service.list(namespace_id=namespace_id, name=name, psql=psql)


@pubsub_router.post("/namespaces/{namespace_id}/topics", response_model=TopicRead)
async def create_topics(
    namespace_id: UUID4,
    topic_create: TopicCreate = Body(...),
    psql: AsyncSession = Depends(psql_db),
) -> Topic:
    """Create a topic in a namespace if it does not exist.

    Args:
        namespace_id (UUID4): The namespace id.
        topic_create (TopicCreate): The topic to create. Defaults to Body(...).

    Returns:
        Topic: The created topic.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    topic_create.namespace_id = namespace_id
    return await topics_service.create(topic_create=topic_create, psql=psql)


@pubsub_router.delete(
    "/namespaces/{namespace_id}/topics/{topic_id}", response_model=TopicRead
)
async def delete_topics(
    namespace_id: UUID4, topic_id: UUID4, psql: AsyncSession = Depends(psql_db)
) -> Topic:
    """Delete a topic in a namespace.

    Args:
        namespace_id (UUID4): The namespace id.
        topic_id (UUID4): The topic id.

    Returns:
        Topic: The deleted topic.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    topic = await topics_service.get(
        topic_id=topic_id, namespace_id=namespace_id, psql=psql
    )
    if topic is None:
        raise HTTPException(status_code=400, detail="Topic not found.")
    return await topics_service.delete(
        topic_id=topic_id, namespace_id=namespace_id, psql=psql
    )


@pubsub_router.post(
    "/namespaces/{namespace_id}/topics/{topic_id}/subscriptions",
    response_model=SubscriptionRead,
)
async def create_subscriptions(
    namespace_id: UUID4,
    topic_id: UUID4,
    subscription_create: SubscriptionCreate = Body(...),
    psql: AsyncSession = Depends(psql_db),
) -> Subscription:
    """Create a subscription to a topic in a namespace if it does not exist.

    Args:
        namespace_id (UUID4): The namespace id.
        topic_id (UUID4): The topic id.
        subscription_create (SubscriptionCreate): The subscription to create.
            Defaults to Body(...).

    Returns:
        Subscription: The created subscription.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    topic = await topics_service.get(
        topic_id=topic_id, psql=psql, namespace_id=namespace_id
    )
    if topic is None:
        raise HTTPException(status_code=400, detail="Topic not found.")
    return await subscriptions_service.create(
        subscription_create=subscription_create, psql=psql
    )


@pubsub_router.get(
    "/namespaces/{namespace_id}/topics/{topic_id}/subscriptions",
    response_model=List[SubscriptionRead],
)
async def get_subscriptions(
    namespace_id: UUID4,
    topic_id: UUID4,
    name: Optional[str] = None,
    psql: AsyncSession = Depends(psql_db),
) -> List[Subscription]:
    """Get all subscriptions to a topic in a namespace.

    Args:
        namespace_id (UUID4): The namespace id.
        topic_id (UUID4): The topic id.
        name (Optional[str], optional): The subscription name. Defaults to None.

    Returns:
        List[Subscription]: The subscriptions.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    topic = await topics_service.get(
        topic_id=topic_id, psql=psql, namespace_id=namespace_id
    )
    if topic is None:
        raise HTTPException(status_code=400, detail="Topic not found.")
    return await subscriptions_service.list(
        topic_id=topic_id, namespace_id=namespace_id, name=name, psql=psql
    )


@pubsub_router.delete(
    "/namespaces/{namespace_id}/topics/{topic_id}/subscriptions/{subscription_id}",
    response_model=SubscriptionRead,
)
async def delete_subscriptions(
    namespace_id: UUID4,
    topic_id: UUID4,
    subscription_id: UUID4,
    psql: AsyncSession = Depends(psql_db),
) -> Subscription:
    """Delete a subscription to a topic in a namespace.

    Args:
        namespace_id (UUID4): The namespace id.
        topic_id (UUID4): The topic id.
        subscription_id (UUID4): The subscription id.

    Returns:
        Subscription: The deleted subscription.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    topic = await topics_service.get(
        topic_id=topic_id, psql=psql, namespace_id=namespace_id
    )
    if topic is None:
        raise HTTPException(status_code=400, detail="Topic not found.")
    subscription = await subscriptions_service.get(
        subscription_id=subscription_id,
        topic_id=topic_id,
        namespace_id=namespace_id,
        psql=psql,
    )
    if subscription is None:
        raise HTTPException(status_code=400, detail="Subscription not found.")
    return await subscriptions_service.delete(
        subscription_id=subscription_id,
        topic_id=topic_id,
        namespace_id=namespace_id,
        psql=psql,
    )


@pubsub_router.post(
    "/namespaces/{namespace_id}/topics/{topic_id}/publish",
    response_model=None,
)
async def publish_message_to_topic(
    namespace_id: UUID4,
    topic_id: UUID4,
    message: Message = Body(...),
    psql: AsyncSession = Depends(psql_db),
) -> None:
    """Publish a message to a topic in a namespace.

    Args:
        namespace_id (UUID4): The namespace id.
        topic_id (UUID4): The topic id.
        message (Message): The message to publish.

    Returns:
        None.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    topic = await topics_service.get(
        topic_id=topic_id, namespace_id=namespace_id, psql=psql
    )
    if topic is None:
        raise HTTPException(status_code=400, detail="Topic not found.")
    return await topics_service.publish_message(topic=topic, message=message)


@dataset_router.post("/namespaces/{namespace_id}/datasets", response_model=DatasetRead)
async def create_datasets(
    namespace_id: UUID4,
    dataset_create: DatasetCreate = Body(...),
    psql: AsyncSession = Depends(psql_db),
) -> Dataset:
    """Create a dataset in a namespace.

    Args:
        namespace_id (UUID4): The namespace id.
        dataset_create (DatasetCreate): The dataset to create.

    Returns:
        Dataset: The Dataset
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    result = await datasets_service.create(dataset_create=dataset_create, psql=psql)
    # TODO: hudson only supports writing to a dir on a filesystem now
    # hudson will support writing to s3, gcs, huggingface, etc in the future
    await datasets_service.create_directory_for_dataset(dataset=result)
    return result


@dataset_router.get(
    "/namespaces/{namespace_id}/datasets", response_model=List[DatasetRead]
)
async def get_datasets(
    namespace_id: UUID4,
    name: Optional[str] = None,
    psql: AsyncSession = Depends(psql_db),
) -> List[Dataset]:
    """Get all datasets in a namespace.

    Args:
        namespace_id (UUID4): The namespace id.
        name (Optional[str], optional): The name of the dataset. Defaults to None.

    Returns:
        List[Dataset]: The datasets.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    return await datasets_service.list(name=name, psql=psql, namespace_id=namespace_id)


@dataset_router.get(
    "/namespaces/{namespace_id}/datasets/{dataset_id}", response_model=DatasetRead
)
async def get_dataset(
    namespace_id: UUID4, dataset_id: UUID4, psql: AsyncSession = Depends(psql_db)
) -> Dataset:
    """Get a dataset in a namespace.

    Args:
        namespace_id (UUID4): The namespace id.
        dataset_id (UUID4): The dataset id.

    Returns:
        Dataset: The dataset.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    dataset = await datasets_service.get(
        dataset_id=dataset_id, namespace_id=namespace_id, psql=psql
    )
    if dataset is None:
        raise HTTPException(status_code=400, detail="Dataset not found.")
    return dataset


@dataset_router.delete(
    "/namespaces/{namespace_id}/datasets/{dataset_id}", response_model=DatasetRead
)
async def delete_datasets(
    namespace_id: UUID4, dataset_id: UUID4, psql: AsyncSession = Depends(psql_db)
) -> Dataset:
    """Delete a dataset in a namespace.

    Args:
        namespace_id (UUID4): The namespace id.
        dataset_id (UUID4): The dataset id.

    Returns:
        Dataset.
    """
    namespace = await namespace_service.get(namespace_id=namespace_id, psql=psql)
    if namespace is None:
        raise HTTPException(status_code=400, detail="Namespace not found.")
    dataset = await datasets_service.get(
        dataset_id=dataset_id, namespace_id=namespace_id, psql=psql
    )
    if dataset is None:
        raise HTTPException(status_code=400, detail="Dataset not found.")
    return await datasets_service.delete(
        dataset_id=dataset_id, namespace_id=namespace_id, psql=psql
    )


@dataset_router.post(
    "/namespaces/{namespace_id}/datasets/{dataset_id}/write", response_model=None
)
async def write_to_datasets(
    namespace_id: UUID4, dataset_id: UUID4, data_array: DataArray = Body(...)
) -> None:
    """Write to a dataset in a namespace.

    Args:
        namespace_id (UUID4): The namespace id.
        dataset_id (UUID4): The dataset id.
        data_array (DataArray): The data to write.

    Returns:
        None.
    """
    return await datasets_service.write(
        namespace_id=namespace_id, dataset_id=dataset_id, data_array=data_array
    )


@dataset_router.get(
    "/namespaces/{namespace_id}/datasets/{dataset_id}/read", response_model=DataArray
)
async def read_from_datasets(namespace_id: UUID4, dataset_id: UUID4) -> DataArray:
    """Read from a dataset in a namespace.

    Args:
        namespace_id (UUID4): The namespace id.
        dataset_id (UUID4): The dataset id.

    Returns:
        DataArray.
    """
    return await datasets_service.read(
        namespace_id=namespace_id,
        dataset_id=dataset_id,
    )
